/* (c) Victor Yodaiken, Finite State Research LLC. 2016
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Use:
Defines  lock free 1producer/1 consumer fifo queues 
To use with multiple producers add a lock for producers -same with consumers
DEPENDS ON X86 STRONG MEMORY MODEL!! WARNING. 

There are a couple of #defines so that C will do type checking

In C source:
#define OWQ_ENQ my_enq_function
#define OWQ_DEQ my_deq_function
#define OWQ_INIT my_qinit_function
#define OWQ_ELEMENT_T qtype // this is the type of element to be queued

#include "thisfile.h"

Defines

owq_t  <- the type of the queue control structure
static void my_qinit_function( owq_t * q, qtype *a, unsigned int n)
inline static int my_deq_function (owq_t * q, qtype  *i)
inline static int my_enq_function (owq_t * q, qtype  i)


You can have repeats of the defines and include in the same file to
define queue functions for different element types.

The queues are implemented on arrays of OWQ_ELEMENT

A head and tail index are maintained so that enq only increments tail and
deq only increments head which allows producer and consumer to operate
in parallel without write conflicts.

Way too much work was put into distinguishing 2 cases where head==tail using
either the high order bit in head/tail or the low order bit (see end of file)

The high order bit is discarded when head or tail
is used as an index. A deq that fills the queue causes the bit to be
set in head to complement of the bit setting in tail
(which may concurrently change). 
A deq that empties the queue causes the two bits to be set the same.

An alternative, but logically equivalent method is used in the ifdef 0 below.

See the example file for use.
*/
#ifndef OWQ_ENQ 
#define OWQ_ENQ enq_i
#define OWQ_DEQ deq_i
#define OWQ_INIT owq_init_i
#define OWQ_ELEMENT_T  int
#endif

#if 1
#ifndef OWQ_STRUCT_T
typedef struct  {
    unsigned int h, t; void *v; const int z; } owq_t;
#define owqfull(q) ((q->t>>1  == q->h>>1  ) && (q->h != q->t))
#define owqempty(q) ((q->t == q->h))
#define OWQ_STRUCT_T
#endif


#ifndef OWQ_BIT_OFFSET
#define OWQ_BIT_OFFSET ( (sizeof(unsigned int)*8) -1 )
#define OWQ_SETBIT ( 1 << OWQ_BIT_OFFSET ) 
#define OWQ_OFFBIT (~( 1 << OWQ_BIT_OFFSET )) 
#endif

static void OWQ_INIT( owq_t * q, OWQ_ELEMENT_T *a, unsigned int n){

    q->h=q->t = 0;
    *(int *)&q->z = n;
    q->v = (void *) a;
}


inline static int OWQ_DEQ(owq_t * q, OWQ_ELEMENT_T  *i)
{
	int next;
	int h = q->h;
	volatile int t = q->t;
	if (t==h) return -1;
	else {
		h= h& OWQ_OFFBIT;
	*i = ((OWQ_ELEMENT_T *)q->v)[h];
	next  = (h + 1) % q->z;
	if(next  == (t &  OWQ_OFFBIT) ) q->h = t; //empty
	else q->h = next;
	return 0;
    }
}

inline static int OWQ_ENQ(owq_t * q, OWQ_ELEMENT_T i)
{
	int next;
	int t = q->t & OWQ_OFFBIT;
	volatile int h = q->h;
    if ((t == (h& OWQ_OFFBIT))&& (h != q->t)) return -1;
    ((OWQ_ELEMENT_T *)q->v)[t] = i;
    next = (t + 1) % q->z;
    if(next == (h & OWQ_OFFBIT) && ((h & OWQ_SETBIT)== 0))q->t = h| OWQ_SETBIT;
    else q->t = next;

    return 0;

}
#endif
#if 0

#ifndef OWQ_STRUCT_T
typedef struct  {
    unsigned int h, t; void *v; const int z; } owq_t;
#define owqfull(q) ((q->t>>1  == q->h>>1  ) && (q->h != q->t))
#define owqempty(q) ((q->t == q->h))
#define OWQ_STRUCT_T
#endif



static void OWQ_INIT( owq_t * q, OWQ_ELEMENT_T *a, unsigned int n){

    q->h=q->t = 0;
    *(int *)&q->z = n;
    q->v = (void *) a;
}


inline static int OWQ_DEQ(owq_t * q, OWQ_ELEMENT_T  *i)
{
	int next;
	int h = q->h;
	volatile int t = q->t;
	if (t==h) return -1;
	else {
		h=h>>1;
	*i = ((OWQ_ELEMENT_T *)q->v)[h];
	next  = (h + 1) % q->z;
	if(next  == (t>>1) ) q->h = t;
	else q->h = next*2;
	return 0;
    }
}

inline static int OWQ_ENQ(owq_t * q, OWQ_ELEMENT_T i)
{
	int next;
	int t = q->t>>1;
	volatile int h = q->h;
    if ((t == (h>>1))&& (h != q->t)) return -1;
    ((OWQ_ELEMENT_T *)q->v)[t] = i;
    next = (t + 1) % q->z;
    if(next == (h >> 1) && ((h & 1)== 0))q->t = h+1;
    else q->t = next*2;

    return 0;

}
#undef OWQ_ENQ 
#undef OWQ_DEQ 
#undef OWQ_INIT 
#undef OWQ_ELEMENT_T  

#endif

#undef OWQ_ENQ 
#undef OWQ_DEQ 
#undef OWQ_INIT 
#undef OWQ_ELEMENT_T  



